"""
BiCM
========

BiCM is a Python package for computing the Bipartite Configuration Model
of a bipartite network and its statistical projection on one layer.
The paper of reference is at https://www.nature.com/articles/srep10595?origin=ppub

Github repository::

    https://github.com/mat701/BiCM

Author: Matteo Bruno
"""

from .network_functions import *
from .graph_classes import *

__version__ = "2.0.1"
__author__ = """Matteo Bruno (matteo.bruno@imtlucca.it)"""
